(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

theory ITL
imports
   Semantics
begin
                             
section  \<open>Axioms and Rules\<close>

text \<open>
The ITL axiom and proof rules are introduced (taken from 
\cite{moszkowski04:_hierar_compl_proof_propos_inter}) together with the validity operation.
The soundness of the rules and axioms are checked using the lemmas of Semantics.thy.
\<close>

subsection \<open>Rules\<close>

lemma MP :
 assumes "\<turnstile> f \<longrightarrow> g"
         "\<turnstile> f"
 shows   "\<turnstile> g"  
using assms(1) assms(2) by fastforce

lemma BoxGen :
 assumes "\<turnstile> f"
 shows   "\<turnstile> \<box> f"
using assms by (auto simp: always_defs)
    
lemma BiGen:
 assumes "\<turnstile> f"
 shows   "\<turnstile> bi f"
using assms by (auto simp: bi_defs)

subsection \<open>Axioms\<close>
    
lemma ChopAssoc : 
    "\<turnstile>  f ; (g ; h) =  (f;g);h "
using  ChopAssocSem Valid_def by blast
 
lemma OrChopImp :
    "\<turnstile> ( f \<or> g);h   \<longrightarrow>  f;h \<or> g;h  "
using OrChopImpSem Valid_def by blast

lemma ChopOrImp :
    "\<turnstile>  f;(g \<or> h) \<longrightarrow>   f;g \<or>  f;h "
using ChopOrImpSem Valid_def by blast
      
lemma EmptyChop :
    "\<turnstile>  empty ; f = f  "
using EmptyChopSem Valid_def by blast

lemma ChopEmpty :
    "\<turnstile> f;empty = f  "
using ChopEmptySem Valid_def by blast

lemma StateImpBi :
    "\<turnstile> init f \<longrightarrow>   bi (init f)  "
using StateImpBiSem Valid_def by blast

lemma NextImpNotNextNot :
    "\<turnstile>  \<circle> f \<longrightarrow>  \<not> (\<circle> (\<not> f)) "
using NextImpNotNextNotSem Valid_def by blast

lemma BiBoxChopImpChop :
    "\<turnstile>  bi ( f \<longrightarrow> f1) \<and> \<box>(g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f1;g1 "
using BiBoxChopImpChopSem Valid_def by blast

lemma BoxInduct :
    "\<turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f "
using BoxInductSem Valid_def by blast
   
lemma ChopstarEqv :
    "\<turnstile>  f\<^sup>\<star> =  (empty \<or>  (f \<and> more); f\<^sup>\<star>) " 
using ChopstarEqvSem Valid_def by blast

subsection \<open>Quantification\<close>


lemma EExI :
 "\<turnstile> F y \<longrightarrow> (\<exists>\<exists> x . F x)"
by (simp add: exist_state_d_def Valid_def, auto) 


lemma EExE: 
 "\<lbrakk>\<And>x. \<turnstile> F x \<longrightarrow> G\<rbrakk> \<Longrightarrow> \<turnstile> (\<exists>\<exists> x. F x) \<longrightarrow> G "
by (metis (mono_tags, lifting) Valid_def exist_state_d_def unl_lift2)

lemma EExVal: 
  "(w \<Turnstile> (\<exists>\<exists> x. F x)) = 
   (\<exists> x (val :: 'a interval). ( ( val = (map x w) \<and>  (w \<Turnstile> F x))))"
by (simp add: exist_state_d_def)

lemma AAxDef:
 "\<turnstile> (\<forall>\<forall> x. F x) = (\<not>(\<exists>\<exists> x. \<not>(F x)))"
by (simp add: Valid_def forall_state_d_def exist_state_d_def) 

lemma EExRev :
 "\<turnstile> (\<exists>\<exists> x. F x)\<^sup>r = (\<exists>\<exists> x. (F x)\<^sup>r)"
by (simp add: Valid_def exist_state_d_def reverse_d_def)

lemma ExEqvRule:
assumes "\<And> x. \<turnstile> (f x) = (g x)"
  shows "\<turnstile> (\<exists> x. f x) = (\<exists> x. g x)"
using assms by fastforce
  

subsection \<open>Lemmas about \<open>current_val\<close>\<close>

lemma current_const: "\<turnstile> $(#c) = #c"
  by (auto simp: current_val_d_def)

lemma current_fun1: "\<turnstile> $(f<x>) = f <$x>"
  by (auto simp: current_val_d_def)

lemma current_fun2: "\<turnstile> $(f<x,y>) = f <$x,$y>"
  by (auto simp: current_val_d_def)

lemma current_fun3: "\<turnstile> $(f<x,y,z>) = f <$x,$y,$z>"
  by (auto simp: current_val_d_def)

lemma current_forall: "\<turnstile> $(\<forall> x. P x) = (\<forall> x. $(P x))"
  by (auto simp: current_val_d_def)

lemma current_exists: "\<turnstile> $(\<exists> x. P x) = (\<exists> x. $(P x))"
  by (auto simp: current_val_d_def)

lemma current_exists1: "\<turnstile> $(\<exists>! x. P x) = (\<exists>! x. $(P x))"
  by (auto simp: current_val_d_def)

lemmas all_current = current_const current_fun1 current_fun2 current_fun3 
  current_forall current_exists current_exists1

lemmas all_current_unl = all_current[THEN intD]
lemmas all_current_eq = all_current[THEN inteq_reflection]


subsection \<open>Lemmas about \<open>next_val\<close>\<close>

lemma next_const: "\<turnstile> more \<longrightarrow> (#c)$ = #c"
  by (auto simp: next_val_d_def more_defs)

lemma next_fun1: "\<turnstile> more \<longrightarrow> f<x>$ = f<x$>"
  by (auto simp: next_val_d_def more_defs)

lemma next_fun2: "\<turnstile> more \<longrightarrow> f<x,y>$ = f <x$,y$>"
  by (auto simp: next_val_d_def more_defs)

lemma next_fun3: "\<turnstile> more \<longrightarrow> f<x,y,z>$ = f <x$,y$,z$>"
  by (auto simp: next_val_d_def more_defs)

lemma next_forall: "\<turnstile> more \<longrightarrow> (\<forall> x. P x)$ = (\<forall> x. (P x)$)"
  by (auto simp: next_val_d_def)

lemma next_exists: "\<turnstile> more \<longrightarrow> (\<exists> x. P x)$ = (\<exists> x. (P x)$)"
  by (auto simp: next_val_d_def)

lemma next_exists1: "\<turnstile> more \<longrightarrow> (\<exists>! x. P x)$ = (\<exists>! x. (P x)$)"
  by (auto simp: next_val_d_def more_defs)

lemmas all_next = next_const next_fun1 next_fun2 next_fun3 
  next_forall next_exists next_exists1 

lemmas all_next_unl = all_next[THEN intD]


subsection \<open>Lemmas about \<open>fin_val\<close>\<close>

lemma fin_const: "\<turnstile> !(#c) = #c"
  by (auto simp: fin_val_d_def)

lemma fin_fun1: "\<turnstile> !(f<x>) = f <!x>"
  by (auto simp: fin_val_d_def)

lemma fin_fun2: "\<turnstile> !(f<x,y>) = f <!x, !y>"
  by (auto simp: fin_val_d_def)

lemma fin_fun3: "\<turnstile> !(f<x,y,z>) = f <!x,!y,!z>"
  by (auto simp: fin_val_d_def)

lemma fin_forall: "\<turnstile> !(\<forall> x. P x) = (\<forall> x. !(P x))"
  by (auto simp: fin_val_d_def)

lemma fin_exists: "\<turnstile> !(\<exists> x. P x) = (\<exists> x. !(P x))"
  by (auto simp: fin_val_d_def)

lemma fin_exists1: "\<turnstile> !(\<exists>! x. P x) = (\<exists>! x. !(P x))"
  by (auto simp: fin_val_d_def)

lemmas all_fin = fin_const fin_fun1 fin_fun2 fin_fun3 
  fin_forall fin_exists fin_exists1

lemmas all_fin_unl = all_fin[THEN intD]
lemmas all_fin_eq = all_fin[THEN inteq_reflection]


subsection \<open>Lemmas about \<open>penult_val\<close>\<close>

lemma penult_const: "\<turnstile> more \<longrightarrow> (#c)! = #c"
  by (auto simp: penult_val_d_def more_defs)

lemma penult_fun1: "\<turnstile> more \<longrightarrow> f<x>! = f<x!>"
  by (auto simp: penult_val_d_def more_defs)

lemma penult_fun2: "\<turnstile> more \<longrightarrow> f<x,y>! = f <x!,y!>"
  by (auto simp: penult_val_d_def more_defs)

lemma penult_fun3: "\<turnstile> more \<longrightarrow> f<x,y,z>! = f <x!,y!,z!>"
  by (auto simp: penult_val_d_def more_defs)

lemma penult_forall: "\<turnstile> more \<longrightarrow> (\<forall> x. P x)! = (\<forall> x. (P x)!)"
  by (auto simp: penult_val_d_def)

lemma penult_exists: "\<turnstile> more \<longrightarrow> (\<exists> x. P x)! = (\<exists> x. (P x)!)"
  by (auto simp: penult_val_d_def)

lemma penult_exists1: "\<turnstile> more \<longrightarrow> (\<exists>! x. P x)! = (\<exists>! x. (P x)!)"
  by (auto simp: penult_val_d_def more_defs)

lemmas all_penult = penult_const penult_fun1 penult_fun2 penult_fun3 
  penult_forall penult_exists penult_exists1 

lemmas all_penult_unl = all_penult[THEN intD]


subsection \<open>Basic temporal variables properties\<close>

lemma empty_imp_fin_eqv_curr:
 "\<turnstile> empty \<longrightarrow> !v = $v"
by (simp add: Valid_def current_val_d_def empty_defs finval_defs)

lemma skip_imp_fin_eqv_next:
 "\<turnstile> skip \<longrightarrow> !v = v$"
by (simp add: Valid_def skip_defs next_val_d_def finval_defs)

lemma skip_imp_penult_eqv_curr:
 "\<turnstile> skip \<longrightarrow> v! = $v"
by (simp add: Valid_def skip_defs penultval_defs current_val_d_def)


subsection \<open>Time reversal properties\<close>

lemma rev_const :
 "\<turnstile> (#c)\<^sup>r = #c"
by (auto simp: reverse_d_def)

lemma rev_fun1 :
 "\<turnstile> (f<x>)\<^sup>r = f<x\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_fun2:
 "\<turnstile> (f<x,y>)\<^sup>r = f<x\<^sup>r,y\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_fun3:
 "\<turnstile> (f<x,y,z>)\<^sup>r = f<x\<^sup>r,y\<^sup>r,z\<^sup>r>"
by (auto simp: reverse_d_def)

lemma rev_forall:
 "\<turnstile> (\<forall> x. P x)\<^sup>r = (\<forall> x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_exists:
 "\<turnstile> (\<exists> x. P x)\<^sup>r = (\<exists> x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_exists1:
 "\<turnstile> (\<exists>! x. P x)\<^sup>r = (\<exists>! x. (P x)\<^sup>r)"
by (auto simp: reverse_d_def)

lemma rev_current:
 "\<turnstile> ($v)\<^sup>r = (!v)"
by (auto simp: interval_intrev_nth current_val_d_def fin_val_d_def reverse_d_def) 

lemma rev_next:
 "\<turnstile>  (v$)\<^sup>r = (v!)"
by (auto simp: interval_intrev_nth next_val_d_def penult_val_d_def reverse_d_def) 

lemma rev_penult:
 "\<turnstile>  (v!)\<^sup>r = (v$)"
by (auto simp: interval_intrev_nth next_val_d_def penult_val_d_def reverse_d_def) 

lemma rev_fin:
 "\<turnstile> (!v)\<^sup>r = ($v)"
by (auto simp:  interval_intrev_nth fin_val_d_def current_val_d_def reverse_d_def) 

lemma EqvReverseReverse:
 "\<turnstile> (f\<^sup>r)\<^sup>r = f"
by (simp add: Valid_def reverse_d_def)

lemma ReverseEqv:
  "(\<turnstile> f) \<longleftrightarrow> (\<turnstile> f\<^sup>r)"
by (metis Valid_def interval_rev_swap reverse_d_def)

lemma RevSkip:
 "\<turnstile> skip\<^sup>r = skip"
by (simp add: Valid_def reverse_d_def skip_defs)

lemma RevChop:
"\<turnstile> (f;g)\<^sup>r = (g\<^sup>r;f\<^sup>r)"
apply (simp add: Valid_def reverse_d_def chop_d_def)
 using interval_intrev_prefix interval_intrev_suffix 
by (metis diff_diff_cancel diff_le_self)

lemma RMoreEqvMore:
 "\<turnstile> more\<^sup>r = more"
apply (simp add: Valid_def more_d_def next_d_def chop_d_def skip_d_def reverse_d_def)
by (simp add: interval_prefix_length)

lemma REmptyEqvEmpty:
 "\<turnstile> empty\<^sup>r = empty"
by (metis RMoreEqvMore empty_d_def int_eq rev_fun1)

lemma PowerChopCommute:
 "\<turnstile> ((f \<and> more);(powerchop f n)) = (powerchop f n);(f \<and> more)"
proof 
 (induct n)
 case 0
 then show ?case using EmptyChopSem  ChopEmptySem power_0 Valid_def by (metis inteq_reflection)
 next
 case (Suc n)
 then show ?case 
 by (metis ChopAssocSem intI inteq_reflection power_chop_d.power_Suc)
qed

lemma REqvRule:
 assumes "\<turnstile> f =g"
 shows " \<turnstile> (f\<^sup>r) = (g\<^sup>r)"
using assms 
using inteq_reflection by force

lemma RevPowerChop:
 "\<turnstile> (powerchop f n)\<^sup>r = (powerchop (f\<^sup>r) n)"
proof 
 (induct n)
 case 0
 then show ?case using REmptyEqvEmpty by auto
 next
 case (Suc n)
 then show ?case 
 by (metis PowerChopCommute RevChop RMoreEqvMore int_eq power_chop_d.power_Suc rev_fun2)
qed

lemma RevChopstar:
 "\<turnstile> (f\<^sup>\<star>)\<^sup>r = (f\<^sup>r)\<^sup>\<star>"
proof -
 have 1: "\<turnstile> (f\<^sup>\<star>) = (\<exists>n. powerchop f n)" 
      by (simp add: chopstar_eqv_power_chop Valid_def) 
 have 2: "\<turnstile> (f\<^sup>\<star>)\<^sup>r = (\<exists> n. powerchop f n)\<^sup>r " 
      using REqvRule "1" by blast
 have 3: "\<turnstile> (\<exists> n. powerchop f n)\<^sup>r = (\<exists> n. (powerchop f n)\<^sup>r)" 
      by (simp add: rev_exists)
 have 4: "\<turnstile> (\<exists> n. (powerchop f n)\<^sup>r) = (\<exists> n. (powerchop (f\<^sup>r) n))" 
      by (simp add: RevPowerChop ExEqvRule)  
 have 5: "\<turnstile> (\<exists> n. (powerchop (f\<^sup>r) n)) = (f\<^sup>r)\<^sup>\<star>" 
      by (simp add: chopstar_eqv_power_chop Valid_def)
 from 2 3 4 5 show ?thesis by fastforce
qed

lemmas all_rev = rev_const rev_fun1 rev_fun2 rev_fun3 rev_forall rev_exists
  rev_exists1 rev_current rev_next rev_penult rev_fin RevSkip RevChop RevChopstar

lemmas all_rev_unl = all_rev[THEN intD]
lemmas all_rev_eq = all_rev[THEN inteq_reflection]



end